#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <assert.h>
#include <math.h>

#include "Curves.h"
#include "CurveEvaluators.h"
#include "CurveEditor.h"

#include "LinearCurveEvaluator.h"

const double CurveEditor::CONTROL_POINT_SELECTION_TOLERANCE	= 0.025;
const double CurveEditor::CURVE_SELECTION_TOLERANCE			= 0.025;

const float CurveEditor::GRID_BACKGROUND_INTENSITY					= static_cast<float>(0.65);
const float CurveEditor::GRID_LINE_INTENSITY						= static_cast<float>(0.85);
const float CurveEditor::GRID_SHADOW_INTENSITY						= static_cast<float>(0.15);

const unsigned int CurveEditor::NUMBER_OF_TICKS						= 20;

CurveEditor::CurveEditor(void)
			:m_wrap_control_points(false),
			 m_have_incomplete_zoom(false),
			 m_currently_curve_moving_control_point_index_is_valid(false),
			 m_currently_selected_curve(NULL)

{
	CurveEvaluators * curve_evaluators = CurveEvaluators::getInstance();

	curve_evaluators->addCurveEvaluator("Linear", new LinearCurveEvaluator());
	
	// ADD YOUR CODE HERE
	//
	// Before you can use your new curve, you need to register them with the animator.
	// You need to call addCurveEvaluator function to do this.

}

CurveEditor::~CurveEditor(void)
{
}

void CurveEditor::specifyWindowSize(const unsigned int & window_width, const unsigned int & window_height)
{
	m_window_width	= window_width;
	m_window_height	= window_height;

	setupWindow();
}

void CurveEditor::changeAnimationLength(const double & animation_length)
{
	zoomOutTotally();
	
	Curves::getInstance()->changeAnimationLength(animation_length);
}

void CurveEditor::rescaleTime(const double & scale_factor)
{
	zoomOutTotally();

	Curves * curves = Curves::getInstance();

	for (Curves::iterator curve_iterator = curves->begin(); curve_iterator != curves->end(); ++curve_iterator)
	{
		curve_iterator->scaleX(scale_factor);
	}
}

double CurveEditor::getMinimumTime(void) const
{
	double minimum_time = 0.0;

	if (m_currently_selected_curve != NULL)
	{
		minimum_time = m_currently_selected_curve->getZoomer()->getLeft();
	}
	else
	{
		Curves * curves = Curves::getInstance();

		if (curves->getNumberOfCurves() > 0)
		{
			minimum_time = curves->begin()->getZoomer()->getLeft();
		}
	}
	
	return minimum_time;
}

double CurveEditor::getMaximumTime(void) const
{
	double maximum_time = 0.0;

	if (m_currently_selected_curve != NULL)
	{
		maximum_time = m_currently_selected_curve->getZoomer()->getRight();
	}
	else
	{
		Curves * curves = Curves::getInstance();

		if (curves->getNumberOfCurves() > 0)
		{
			maximum_time = curves->begin()->getZoomer()->getRight();
		}
	}
	
	return maximum_time;
}

double CurveEditor::getMinimumValue(void) const
{
	double minimum_value = 0.0;

	if (m_currently_selected_curve != NULL)
	{
		minimum_value = m_currently_selected_curve->getZoomer()->getBottom();
	}
	
	return minimum_value;
}

double CurveEditor::getMaximumValue(void) const
{
	double maximum_value = 0.0;

	if (m_currently_selected_curve != NULL)
	{
		maximum_value = m_currently_selected_curve->getZoomer()->getTop();
	}
	
	return maximum_value;
}

void CurveEditor::drawCurveEditor(const double & current_time) const
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	setupWindow();

	drawHashMarks();

	drawAllCurves();

	drawIncompleteZoom();

	drawTimeBar(current_time);
}

double CurveEditor::getTime(const unsigned int & x_location) const
{
	int clamped_x_location = clamp(x_location, 0, m_window_width - 1);

	double width_fraction	= (static_cast<double>(clamped_x_location) / static_cast<double>(m_window_width));

	assert(width_fraction >= 0.0);
	assert(width_fraction <= 1.0);

	double time = 0.0;

	time = Curves::getInstance()->getTime(width_fraction);

	return time;
}

double CurveEditor::getNormalizedX(const unsigned int & x_location) const
{
	double width_fraction	= (static_cast<double>(x_location) / static_cast<double>(m_window_width));

	assert(width_fraction >= 0.0);
	assert(width_fraction <= 1.0);

	return width_fraction;
}

double CurveEditor::getNormalizedY(const unsigned int & y_location) const
{
	double height_fraction	= 1.0 - (static_cast<double>(y_location) / static_cast<double>(m_window_height));

	assert(height_fraction >= 0.0);
	assert(height_fraction <= 1.0);

	return height_fraction;
}

void CurveEditor::mouseDown(const int & x_location, const int & y_location)
{
	int clamped_x_location = clamp(x_location, 0, m_window_width - 1);
	int clamped_y_location = clamp(y_location, 0, m_window_height - 1);
	
	validateCoordinates(clamped_x_location, clamped_y_location);

	if (m_currently_selected_curve != NULL)
	{
		Point location(getNormalizedX(clamped_x_location), getNormalizedY(clamped_y_location));

		Curve * closest_curve = NULL;

		unsigned int closest_curve_control_point_index;

		double closest_curve_control_point_distance;

		closest_curve_control_point_index = m_currently_selected_curve->getClosestControlPoint(location, &closest_curve_control_point_distance);

		if (closest_curve_control_point_distance <= CONTROL_POINT_SELECTION_TOLERANCE)
		{
			m_currently_curve_moving_control_point_index = closest_curve_control_point_index;

			m_currently_curve_moving_control_point_index_is_valid = true;
		}
		else
		{
			m_currently_selected_curve->addControlPoint(location);

			double zero_distance;

			m_currently_curve_moving_control_point_index = m_currently_selected_curve->getClosestControlPoint(location, &zero_distance);

			m_currently_curve_moving_control_point_index_is_valid = true;

			static const double EPSILON = 0.02;

			assert(zero_distance < EPSILON);
		}
	}
}

void CurveEditor::altClick(const int & x_location, const int & y_location)
{
	int clamped_x_location = clamp(x_location, 0, m_window_width - 1);
	int clamped_y_location = clamp(y_location, 0, m_window_height - 1);
	
	validateCoordinates(clamped_x_location, clamped_y_location);

	Point location(getNormalizedX(clamped_x_location), getNormalizedY(clamped_y_location));

	Curve * closest_curve = NULL;

	double closest_curve_control_point_distance;

	Curves * curves = Curves::getInstance();

	curves->getClosestShownCurve(location, &closest_curve_control_point_distance, &closest_curve);

	if (closest_curve_control_point_distance <= CURVE_SELECTION_TOLERANCE)
	{
		m_currently_selected_curve = closest_curve;
	}
	else
	{
		m_currently_selected_curve = NULL;
	}
}

void CurveEditor::controlClick(const int & x_location, const int & y_location)
{
	int clamped_x_location = clamp(x_location, 0, m_window_width - 1);
	int clamped_y_location = clamp(y_location, 0, m_window_height - 1);
	
	validateCoordinates(clamped_x_location, clamped_y_location);

	Curves * curves = Curves::getInstance();

	if (curves->getNumberOfCurves() > 0)
	{
		Point location(getNormalizedX(clamped_x_location), getNormalizedY(clamped_y_location));

		Curve * closest_curve = NULL;

		unsigned int closest_curve_control_point_index;

		double closest_curve_control_point_distance;

		curves->getClosestControlPointOnShownCurve(location, &closest_curve_control_point_distance, &closest_curve, &closest_curve_control_point_index);

		if (closest_curve != NULL)
		{
			if (closest_curve->getNumberOfControlPoints() > 2)
			{
				if (closest_curve_control_point_distance <= CONTROL_POINT_SELECTION_TOLERANCE)
				{
					closest_curve->removeControlPoint(closest_curve_control_point_index);
				}
			}
		}
	}

	m_currently_curve_moving_control_point_index_is_valid = false;
}

void CurveEditor::mouseMove(const int & x_location, const int & y_location)
{
	int clamped_x_location = clamp(x_location, 0, m_window_width - 1);
	int clamped_y_location = clamp(y_location, 0, m_window_height - 1);
	
	validateCoordinates(clamped_x_location, clamped_y_location);
	
	if (m_currently_selected_curve != NULL)
	{
		Point new_location(getNormalizedX(clamped_x_location), getNormalizedY(clamped_y_location));

		if (m_currently_curve_moving_control_point_index_is_valid)
		{
			m_currently_selected_curve->changeControlPoint(m_currently_curve_moving_control_point_index, new_location);
		}
	}
}

void CurveEditor::mouseUp(const int & x_location, const int & y_location)
{
	int clamped_x_location = clamp(x_location, 0, m_window_width - 1);
	int clamped_y_location = clamp(y_location, 0, m_window_height - 1);
	
	validateCoordinates(clamped_x_location, clamped_y_location);

	m_currently_curve_moving_control_point_index_is_valid = false;
}

void CurveEditor::shiftMouseDown(const int & x_location, const int & y_location)
{
	int clamped_x_location = clamp(x_location, 0, m_window_width - 1);
	int clamped_y_location = clamp(y_location, 0, m_window_height - 1);
	
	validateCoordinates(clamped_x_location, clamped_y_location);

	Curves * curves = Curves::getInstance();
	
	double normalized_x = getNormalizedX(clamped_x_location);
	double normalized_y = getNormalizedY(clamped_y_location);

	m_incomplete_zoom.setLeftBottom(normalized_x, normalized_y);
	m_incomplete_zoom.setRightTop  (normalized_x, normalized_y);

	m_have_incomplete_zoom = true;

	curves->startZoomSelection(normalized_x, normalized_y);
	curves->finishZoomSelection(normalized_x, normalized_y);
}

void CurveEditor::shiftMouseDrag(const int & x_location, const int & y_location)
{
	int clamped_x_location = clamp(x_location, 0, m_window_width - 1);
	int clamped_y_location = clamp(y_location, 0, m_window_height - 1);
	
	validateCoordinates(clamped_x_location, clamped_y_location);

	Curves * curves = Curves::getInstance();
	
	double normalized_x = getNormalizedX(clamped_x_location);
	double normalized_y = getNormalizedY(clamped_y_location);

	m_incomplete_zoom.setRightTop  (normalized_x, normalized_y);

	m_have_incomplete_zoom = true;

	curves->finishZoomSelection(normalized_x, normalized_y);
}

void CurveEditor::shiftMouseUp(const int & x_location, const int & y_location)
{
	int clamped_x_location = clamp(x_location, 0, m_window_width - 1);
	int clamped_y_location = clamp(y_location, 0, m_window_height - 1);
	
	validateCoordinates(clamped_x_location, clamped_y_location);

	Curves * curves = Curves::getInstance();
	
	double normalized_x = getNormalizedX(clamped_x_location);
	double normalized_y = getNormalizedY(clamped_y_location);

	m_incomplete_zoom.setRightTop  (normalized_x, normalized_y);

	m_have_incomplete_zoom = false;

	curves->finishZoomSelection(normalized_x, normalized_y);
	
	curves->pushZoom();
}

void CurveEditor::tryPopZoom(void)
{
	Curves * curves = Curves::getInstance();

	if (curves->canZoomOut())
	{
		curves->popZoom();
	}
}

void CurveEditor::zoomOutTotally(void)
{
	Curves * curves = Curves::getInstance();

	while (curves->canZoomOut())
	{
		curves->popZoom();
	}
}

void CurveEditor::changeMinimumY(const double & new_minimum_y)
{
	if (m_currently_selected_curve != NULL)
	{
		zoomOutTotally();
		
		Zoomer * zoom = m_currently_selected_curve->getZoomer();

		zoom->setupObjectSpaceDimensions(zoom->getLeft(), zoom->getRight(), new_minimum_y, zoom->getTop());
	}
}

void CurveEditor::changeMaximumY(const double & new_maximum_y)
{
	if (m_currently_selected_curve != NULL)
	{
		zoomOutTotally();

		Zoomer * zoom = m_currently_selected_curve->getZoomer();

		zoom->setupObjectSpaceDimensions(zoom->getLeft(), zoom->getRight(), zoom->getBottom(), new_maximum_y);
	}
}

Curve * CurveEditor::getCurrentlySelectedCurve(void) const
{
	return m_currently_selected_curve;
}

void CurveEditor::toggleShowAnimationCurve(const std::string & name)
{
	Curve * curve = Curves::getInstance()->getCurve(name);
		
	curve->toggleShowCurve();

	m_currently_selected_curve = curve;
}

void CurveEditor::toggleWrapControlPoints(void)
{
	m_wrap_control_points = ! m_wrap_control_points;

	Curves::getInstance()->setWrapControlPointsFlags(m_wrap_control_points);
}

// private member functions

void CurveEditor::setupWindow(void) const
{
	glMatrixMode(GL_PROJECTION);

	glLoadIdentity();

	glMatrixMode(GL_MODELVIEW);

	glLoadIdentity();

	glViewport(0, 0, m_window_width, m_window_height);

	glDisable(GL_DEPTH_TEST);
	glDrawBuffer(GL_BACK);

	glClearColor(GRID_BACKGROUND_INTENSITY, GRID_BACKGROUND_INTENSITY, GRID_BACKGROUND_INTENSITY, 0.0);
}

void CurveEditor::drawAllCurves(void) const
{
	Curves::getInstance()->drawAllCurves();

	if (m_currently_selected_curve != NULL)
	{
		glColor3d(1.0, 1.0, 1.0);

		m_currently_selected_curve->getZoomer()->applyTopZoom();

		m_currently_selected_curve->drawEvaluatedCurveSegments();

		glPointSize(5.0);

			m_currently_selected_curve->drawControlPoints();

		glPointSize(1.0);
	}

	glFlush();
}

void CurveEditor::findNiceDivisions(const double & minimum, const double & maximum, double * new_start, double * new_division_interval, double * new_end) const
{
	// stolen without shame from Doug's code
	// I don't know what "w" is off the top of my head.
	
	double w = (maximum - minimum) / 10.0;

	if (w != 0.0)
	{
		int exponent = 0;

		while (w < 1.0)
		{
			exponent--;

			w *= 10.0;
		}

		while (w > 10.0)
		{
			exponent++;

			w /= 10.0;
		}

		int i = static_cast<unsigned int>(w);

		if ((i > 2) && (i < 5))
		{
			i = 2;
		}
		else if (i > 5)
		{
			i = 5;
		}

		double division_interval = static_cast<double>(i) * pow(10.0, exponent);

		double start;

		if (minimum >= 0.0)
		{
			start = minimum - fmod(minimum, division_interval) + division_interval;
		}
		else
		{
			start = minimum - fmod(minimum, division_interval);

			if (start == minimum)
			{
				start += division_interval;
			}
		}

		*new_start				= start;
		*new_division_interval	= division_interval;

		*new_end = *new_start;

		while ((*new_end + *new_division_interval) <= maximum)
		{
			*new_end += *new_division_interval;
		}
	}
	else
	{
		*new_start				= 0.0;
		*new_division_interval	= 0.0;
	}
}

void CurveEditor::drawHashMarks(void) const
{
	glMatrixMode(GL_PROJECTION);

	glPushMatrix();

	glLoadIdentity();

	gluOrtho2D(0, m_window_width, 0, m_window_height);

	glBegin(GL_LINES);

		double width_tick_start;
		double width_tick_division_interval;
		double width_tick_end;

		findNiceDivisions(getMinimumTime(), getMaximumTime(), &width_tick_start, &width_tick_division_interval, &width_tick_end);

		unsigned int width_tick_interval = static_cast<unsigned int>(static_cast<double>(m_window_width) * width_tick_division_interval / (getMaximumTime() - getMinimumTime()));

		if (width_tick_interval != 0)
		{
			unsigned int width_tick_start_int = static_cast<unsigned int>(static_cast<double>(m_window_width) * (width_tick_start - getMinimumTime()) / (getMaximumTime() - getMinimumTime()));

			for (unsigned int width_tick_location = width_tick_start_int ; width_tick_location < m_window_width; width_tick_location += width_tick_interval)
			{
				glColor3d(GRID_LINE_INTENSITY, GRID_LINE_INTENSITY, GRID_LINE_INTENSITY);

				glVertex2i(width_tick_location, 0);
				glVertex2i(width_tick_location, m_window_height);

				glColor3d(GRID_SHADOW_INTENSITY, GRID_SHADOW_INTENSITY, GRID_SHADOW_INTENSITY);

				glVertex2i(width_tick_location + 1, 0);
				glVertex2i(width_tick_location + 1, m_window_height);
			}
		}

		double height_tick_start;
		double height_tick_division_interval;
		double height_tick_end;

		findNiceDivisions(getMinimumValue(), getMaximumValue(), &height_tick_start, &height_tick_division_interval, &height_tick_end);

		unsigned int height_tick_interval = static_cast<unsigned int>(static_cast<double>(m_window_height) * height_tick_division_interval / (getMaximumValue() - getMinimumValue()));

		if (height_tick_interval != 0)
		{
			unsigned int height_tick_start_int = static_cast<unsigned int>(static_cast<double>(m_window_height) * (height_tick_start - getMinimumValue()) / (getMaximumValue() - getMinimumValue()));
			
			for (unsigned int height_tick_location = height_tick_start_int; height_tick_location < m_window_height; height_tick_location += height_tick_interval)
			{
				glColor3d(GRID_LINE_INTENSITY, GRID_LINE_INTENSITY, GRID_LINE_INTENSITY);

				glVertex2i(0				, height_tick_location);
				glVertex2i(m_window_width	, height_tick_location);

				glColor3d(GRID_SHADOW_INTENSITY, GRID_SHADOW_INTENSITY, GRID_SHADOW_INTENSITY);

				glVertex2i(0				, height_tick_location + 1);
				glVertex2i(m_window_width	, height_tick_location + 1);
			}
		}

	glEnd();

	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
}

void CurveEditor::drawIncompleteZoom(void) const
{
	if (m_have_incomplete_zoom)
	{
		glMatrixMode(GL_PROJECTION);

		glPushMatrix();

		glLoadIdentity();

		gluOrtho2D(0.0, 1.0, 0.0, 1.0);

		glColor3d(1.0, 1.0, 1.0);

		glBegin(GL_LINE_LOOP);

			glVertex2d(m_incomplete_zoom.getLeft()	, m_incomplete_zoom.getBottom());
			glVertex2d(m_incomplete_zoom.getLeft()	, m_incomplete_zoom.getTop());
			glVertex2d(m_incomplete_zoom.getRight()	, m_incomplete_zoom.getTop());
			glVertex2d(m_incomplete_zoom.getRight()	, m_incomplete_zoom.getBottom());

		glEnd();

		glPopMatrix();

		glMatrixMode(GL_MODELVIEW);
	}
}

void CurveEditor::drawTimeBar(const double & current_time) const
{
	Curves * curves = Curves::getInstance();

	if (curves->getNumberOfCurves() > 0)
	{
		glColor3d(1.0, 0.0, 0.0);

		glMatrixMode(GL_PROJECTION);

		glPushMatrix();

		glLoadIdentity();

		Curve * arbitrary_first_curve = curves->begin();

		Zoomer * zoom = arbitrary_first_curve->getZoomer();

		gluOrtho2D(zoom->getLeft(), zoom->getRight(), 0, 1);

		glLineWidth(5.0);

		glBegin(GL_LINES);

			glVertex2d(current_time, 0.0);
			glVertex2d(current_time, 1.0);

		glEnd();

		glLineWidth(1.0);

		glPopMatrix();

		glMatrixMode(GL_MODELVIEW);
	}
}

int CurveEditor::clamp(const int & x, const int & a, const int & b) const
{
	int new_x;

	if (x < a)
	{
		new_x = a;
	}
	else if (x > b)
	{
		new_x = b;
	}
	else
	{
		new_x = x;
	}

	assert(new_x >= a);
	assert(new_x <= b);

	return new_x;
}

void CurveEditor::validateCoordinates(const int & x_location, const int & y_location) const
{
	assert (x_location >= 0);
	assert (x_location < m_window_width);

	assert (y_location >= 0);
	assert (y_location < m_window_height);
}
